// evt.cxx

#include <stdio.h>
#include <tchar.h>

#define _WIN32_WINNT 0x0500
#include <windows.h>

#include "winnthook.h"

#include "resource.h"

#define WM_MSGBACKAGAIN		( WM_USER + 7 )

static unsigned uKbdKeyCount;
static unsigned uKbdSysKeyCount;
static unsigned uMouseLeftCount;
static unsigned uMouseRghtCount;
static unsigned uMouseMiddleCount;
static unsigned uMouseOthrCount;
static unsigned uMouseWheelCount;
static unsigned long ulMouseMoveDistance;

static BOOL boolHideWhenMinimized;
static BOOL boolIsHidden;

static SYSTEMTIME systimeStart;

static HINSTANCE hInstance;

static HICON hiconApp;

static HACCEL haccelMain;

static HMENU hmenuMain;

static HWND hwndMain;

static HWND hwndKeybrdHookEdit;
static HWND hwndMouseHookEdit;

__declspec( dllimport ) void __stdcall SetKeybrdHookHlpr( void (CALLBACK * pfnKeybrdHookHlprArg) ( WPARAM, LPARAM ) );
__declspec( dllimport ) void __stdcall SetMouseHookHlpr( void (CALLBACK * pfnMouseHookHlprArg) ( WPARAM, LPARAM ) );

__declspec( dllimport ) int __stdcall InitHooks( void );
__declspec( dllimport ) int __stdcall DoneHooks( void );

__declspec( dllimport ) int __stdcall GetInstanceCount( void );

void CALLBACK KeybrdHookHlpr( WPARAM wParam, LPARAM lParam );
void CALLBACK MouseHookHlpr( WPARAM wParam, LPARAM lParam );

BOOL CALLBACK MainFrameProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

BOOL CALLBACK CounterDetailsDlgProc( HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam );

unsigned long __stdcall CalcDistance( POINT& ptDest, const POINT& ptSrc );

void __stdcall ResetCounter( void );

void __stdcall BackAgainApp( void );
void __stdcall HideApp( void );
void __stdcall NotificationSymbol( DWORD dwMsg = (DWORD) NIM_MODIFY );

int APIENTRY _tWinMain( HINSTANCE hInstance, HINSTANCE /*hPrevInstance*/, LPTSTR /*lpsztCmdLine*/, int nCmdShow )
{
	static _TCHAR sztClassName[ 130 ];
	static _TCHAR sztErrMsg[ 130 ];
	static WNDCLASS wc;

	int nInstanceCount = (int) GetInstanceCount( );
	int nRet = (int) LoadString( ::hInstance = hInstance, (UINT) IDR_MAINFRAME, (LPTSTR) sztClassName, 128 );

	if ( nInstanceCount > 1 )
	{
		_stprintf( sztErrMsg, TEXT( "Only 1 instance is allowed or possible.\n(count=%d)." ), nInstanceCount );
		goto ErrMsg;
	}

	if ( ( nRet <= 0 ) || ( nRet > 128 ) )
	{
		_stprintf( sztErrMsg, TEXT( "Unable to get the main dialog class." ) );
		goto ErrMsg;
	}

	hiconApp = (HICON) LoadIcon( (HINSTANCE) hInstance, MAKEINTRESOURCE( IDR_MAINFRAME ) );

	wc.style = CS_CLASSDC;
	wc.lpfnWndProc = (WNDPROC) MainFrameProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = DLGWINDOWEXTRA;
	wc.hInstance = hInstance;
	wc.hIcon = hiconApp;
	wc.hCursor = LoadCursor( NULL, IDC_ARROW );
	wc.hbrBackground = (HBRUSH) COLOR_WINDOW;
	wc.lpszMenuName  = (LPCTSTR) NULL;
	wc.lpszClassName = (LPCTSTR) sztClassName;

	if ( RegisterClass( (const WNDCLASS*) &wc ) );
	else
	{
		_stprintf( sztErrMsg, TEXT( "API call RegisterClass ('%s') failed." ), sztClassName );
		DestroyIcon( hiconApp );
		goto ErrMsg;
	}

	haccelMain = (HACCEL) LoadAccelerators( hInstance, MAKEINTRESOURCE( IDR_MAINFRAME ) );

	hwndMain = CreateDialog( hInstance, MAKEINTRESOURCE( IDR_MAINFRAME ), (HWND) NULL, (DLGPROC) NULL );

	ShowWindow( hwndMain, nCmdShow );
	UpdateWindow( hwndMain );

	hmenuMain = GetMenu( hwndMain );

	hwndKeybrdHookEdit = GetDlgItem( hwndMain, IDC_KEYBRDHOOKEDIT );
	hwndMouseHookEdit = GetDlgItem( hwndMain, IDC_MOUSEHOOKEDIT );

	SetKeybrdHookHlpr( KeybrdHookHlpr );
	SetMouseHookHlpr( MouseHookHlpr );

	MSG msg;

	while ( GetMessage( (LPMSG) &msg, (HWND) NULL, (UINT) 0, (UINT) 0 ) )
	{
		TranslateAccelerator( (HWND) hwndMain, (HACCEL) haccelMain, (LPMSG) &msg );
		if ( IsDialogMessage( (HWND) hwndMain, (LPMSG) &msg ) );
		else
		{
			TranslateMessage( (CONST MSG*) &msg );
			DispatchMessage( (CONST MSG*) &msg );
		}
	}

	DestroyAcceleratorTable( haccelMain );
	if ( UnregisterClass( (LPCTSTR) sztClassName, hInstance ) );
	else
	{
		_stprintf( sztErrMsg, TEXT( "API call UnregisterClass ('%s') failed." ), sztClassName );
		DestroyIcon( hiconApp );
		goto ErrMsg;
	}
	DestroyIcon( hiconApp );

	return 0;

ErrMsg:	MessageBox( (HWND) NULL, sztErrMsg, (LPCTSTR) TEXT( "Event-Counter Utility" ), (UINT) ( MB_OK | MB_ICONERROR | MB_TASKMODAL ) );

	return -1;
}

void CALLBACK KeybrdHookHlpr( WPARAM wParam, LPARAM lParam )
{
	PKBDLLHOOKSTRUCT pkbdhk = (PKBDLLHOOKSTRUCT) lParam;

	_TCHAR sztCode[ 20 ], sztKey[ 20 ], sztLine[ 130 ];

	switch ( wParam )
	{
		case (UINT) WM_KEYDOWN :
			_tcscpy( sztCode, TEXT( "KEY DOWN" ) );
			uKbdKeyCount++;
			break;
		case (UINT) WM_SYSKEYDOWN :
			_tcscpy( sztCode, TEXT( "SYSKEY DOWN" ) );
			uKbdSysKeyCount++;
			break;
		case (UINT) WM_KEYUP :
		case (UINT) WM_SYSKEYUP :
			_tcscpy( sztCode, TEXT( "(SYS)KEY UP" ) );
			break;
		default :
			_tcscpy( sztCode, TEXT( "N/A" ) );
	}

	switch ( pkbdhk->vkCode & 0xFF )
	{
		case VK_BACK :
			_tcscpy( sztKey, TEXT( "[back]" ) );
			break;
		case VK_TAB :
			_tcscpy( sztKey, TEXT( "[tab]" ) );
			break;
		case VK_RETURN :
			_tcscpy( sztKey, TEXT( "[enter]" ) );
			break;
		case VK_SHIFT :
			_tcscpy( sztKey, TEXT( "[shift]" ) );
			break;
		case VK_CONTROL :
			_tcscpy( sztKey, TEXT( "[ctrl]" ) );
			break;
		case VK_MENU :
			_tcscpy( sztKey, TEXT( "[alt]" ) );
			break;
		case VK_PAUSE :
			_tcscpy( sztKey, TEXT( "[pause]" ) );
			break;
		case VK_CAPITAL :
			_tcscpy( sztKey, TEXT( "[caps lock]" ) );
			break;
		case VK_ESCAPE :
			_tcscpy( sztKey, TEXT( "[esc]" ) );
			break;
		case VK_SPACE :
			_tcscpy( sztKey, TEXT( "[space]" ) );
			break;
		case VK_PRIOR :
			_tcscpy( sztKey, TEXT( "[page up]" ) );
			break;
		case VK_NEXT :
			_tcscpy( sztKey, TEXT( "[page down]" ) );
			break;
		case VK_END :
			_tcscpy( sztKey, TEXT( "[end]" ) );
			break;
		case VK_HOME :
			_tcscpy( sztKey, TEXT( "[home]" ) );
			break;
		case VK_LEFT :
			_tcscpy( sztKey, TEXT( "[left]" ) );
			break;
		case VK_UP :
			_tcscpy( sztKey, TEXT( "[up]" ) );
			break;
		case VK_RIGHT :
			_tcscpy( sztKey, TEXT( "[right]" ) );
			break;
		case VK_DOWN :
			_tcscpy( sztKey, TEXT( "[down]" ) );
			break;
		case VK_SNAPSHOT :
			_tcscpy( sztKey, TEXT( "[print screen]" ) );
			break;
		case VK_INSERT :
			_tcscpy( sztKey, TEXT( "[insert]" ) );
			break;
		case VK_DELETE :
			_tcscpy( sztKey, TEXT( "[delete]" ) );
			break;
		case VK_LWIN :
			_tcscpy( sztKey, TEXT( "[left win]" ) );
			break;
		case VK_RWIN :
			_tcscpy( sztKey, TEXT( "[right win]" ) );
			break;
		case VK_LSHIFT :
			_tcscpy( sztKey, TEXT( "[left shift]" ) );
			break;
		case VK_RSHIFT :
			_tcscpy( sztKey, TEXT( "[right shift]" ) );
			break;
		case VK_NUMLOCK :
			_tcscpy( sztKey, TEXT( "[num lock]" ) );
			break;
		case VK_SCROLL :
			_tcscpy( sztKey, TEXT( "[scroll lock]" ) );
			break;
		case VK_LMENU :
			_tcscpy( sztKey, TEXT( "[left alt]" ) );
			break;
		case VK_RMENU :
			_tcscpy( sztKey, TEXT( "[right alt]" ) );
			break;
		case VK_MULTIPLY :
			_tcscpy( sztKey, TEXT( "[\x00D7]" ) );
			break;
		case VK_ADD :
			_tcscpy( sztKey, TEXT( "[+]" ) );
			break;
		case VK_SUBTRACT :
			_tcscpy( sztKey, TEXT( "[-]" ) );
			break;
		case VK_DIVIDE :
			_tcscpy( sztKey, TEXT( "[\x00F7]" ) );
			break;
		case VK_DECIMAL :
			_tcscpy( sztKey, TEXT( "[.]" ) );
			break;
		default :
			if ( ( ( pkbdhk->vkCode & 0xFF ) >= VK_F1 ) && ( ( pkbdhk->vkCode & 0xFF ) <= VK_F24 ) )
				_stprintf( sztKey, TEXT( "[F%hd]" ), ( pkbdhk->vkCode & 0xFF ) - VK_F1 + 1 );
			else
				if ( ( ( pkbdhk->vkCode & 0xFF ) >= VK_NUMPAD0 ) && ( ( pkbdhk->vkCode & 0xFF ) <= VK_NUMPAD9 ) )
					_stprintf( sztKey, TEXT( "[%c]" ), ( pkbdhk->vkCode & 0xFF ) - VK_NUMPAD0 + 0x30 );
				else
					if ( ( ( pkbdhk->vkCode & 0xFF ) >= 0x30 ) && ( ( pkbdhk->vkCode & 0xFF ) <= 0x39 ) )
						_stprintf( sztKey, TEXT( "'%c'" ), pkbdhk->vkCode & 0xFF );
					else
						if ( ( ( pkbdhk->vkCode & 0xFF ) >= 0x41 ) && ( ( pkbdhk->vkCode & 0xFF ) <= 0x5A ) )
							_stprintf( sztKey, TEXT( "'%c'" ), pkbdhk->vkCode & 0xFF );
						else
							_stprintf( sztKey, TEXT( "0x%02hx" ), pkbdhk->vkCode & 0xFF );
	}

	_stprintf( sztLine, TEXT( "%s; %s, 0x%04x" ), sztCode, sztKey, pkbdhk->scanCode & 0xFFFF );

	SetWindowText( hwndKeybrdHookEdit, (LPCTSTR) sztLine );
	UpdateWindow( hwndKeybrdHookEdit );

	if ( boolIsHidden )
		NotificationSymbol( );
}

void CALLBACK MouseHookHlpr( WPARAM wParam, LPARAM lParam )
{
	PMSLLHOOKSTRUCT pmshk = (PMSLLHOOKSTRUCT) lParam;

	_TCHAR sztCode[ 20 ], sztLine[ 130 ];

	static POINT pt;

	switch ( wParam )
	{
		case (UINT) WM_LBUTTONDOWN :
			_tcscpy( sztCode, TEXT( "LEFT DOWN" ) );
			uMouseLeftCount++;
			break;
		case (UINT) WM_RBUTTONDOWN :
			_tcscpy( sztCode, TEXT( "RIGHT DOWN" ) );
			uMouseRghtCount++;
			break;
		case (UINT) WM_MBUTTONDOWN :
			_tcscpy( sztCode, TEXT( "MIDDLE(WHEEL) DOWN" ) );
			uMouseMiddleCount++;
			break;
		case (UINT) WM_XBUTTONDOWN :
			_tcscpy( sztCode, TEXT( "OTHER DOWN" ) );
			uMouseOthrCount++;
			break;
		case (UINT) WM_LBUTTONUP :
		case (UINT) WM_RBUTTONUP :
		case (UINT) WM_MBUTTONUP :
		case (UINT) WM_XBUTTONUP :
			_tcscpy( sztCode, TEXT( "ANY UP" ) );
			break;
		case (UINT) WM_MOUSEMOVE :
			_tcscpy( sztCode, TEXT( "MOVE" ) );
			ulMouseMoveDistance += CalcDistance( pt, pmshk->pt );
			break;
		case (UINT) WM_MOUSEWHEEL :
			_tcscpy( sztCode, TEXT( "WHEEL" ) );
			uMouseWheelCount++;
			break;
		default :
			_tcscpy( sztCode, TEXT( "N/A" ) );
	}

	_stprintf( sztLine, TEXT( "%s; (%ld,%ld)" ), sztCode, pmshk->pt.x, pmshk->pt.y );

	SetWindowText( hwndMouseHookEdit, (LPCTSTR) sztLine );
	UpdateWindow( hwndMouseHookEdit );

	if ( boolIsHidden )
		NotificationSymbol( );
}

BOOL CALLBACK MainFrameProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch ( uMsg )
	{
		case (UINT) WM_COMMAND :
			switch ( LOWORD( wParam ) )
			{
				case (WORD) ID_COUNTER_DETAILS :
					if ( DialogBox( hInstance, MAKEINTRESOURCE( IDD_COUNTERDETAILSDIALOG ), hwndMain, (DLGPROC) CounterDetailsDlgProc ) == IDOK )
						if ( MessageBox( hwndMain, (LPCTSTR) TEXT( "Reset counter(s) and runtime?" ), (LPCTSTR) TEXT( "Event-Counter Utility" ), (UINT) ( MB_YESNO | MB_ICONQUESTION | MB_TASKMODAL ) ) == IDYES )
							ResetCounter( );
					break;
				case (WORD) ID_HIDE_WHEN_MINIMIZED :
					if ( boolHideWhenMinimized = ( boolHideWhenMinimized == (BOOL) FALSE ) )
						CheckMenuItem( hmenuMain, (UINT) ID_HIDE_WHEN_MINIMIZED, (UINT) MF_CHECKED );
					else
						CheckMenuItem( hmenuMain, (UINT) ID_HIDE_WHEN_MINIMIZED, (UINT) MF_UNCHECKED );
					break;
				case (WORD) ID_APP_ABOUT :
					ShellAbout( hwndMain, (LPCTSTR) TEXT( "Event-Counter Utility" ), ( LOBYTE( GetVersion( ) ) <= 0x05 ) ? (LPCTSTR) TEXT( " Portions Copyright (C) 2021 Ing. Walter H." ) : (LPCTSTR) TEXT( " \x00A9 Ing. Walter H. All rights resvd." ), (HICON) hiconApp );
					break;
				case (WORD) ID_APP_EXIT :
					PostMessage( hwndMain, (UINT) WM_SYSCOMMAND, (WPARAM) SC_CLOSE, (LPARAM) 0 );
					break;
				default :
					return DefWindowProc( hwnd, uMsg, wParam, lParam );
			}
			break;

		case (UINT) WM_CREATE :
			ResetCounter( );
			InitHooks( );
			break;

		case (UINT) WM_CLOSE :
			DestroyWindow( hwndMain );
			break;

		case (UINT) WM_DESTROY :
			DoneHooks( );
			PostQuitMessage( 0 );
			return FALSE;
			break;

		case (UINT) WM_SYSCOMMAND :
			switch ( LOWORD( wParam ) )
			{
				case (WORD) SC_MINIMIZE :
					if ( boolHideWhenMinimized )
						HideApp( );
					else
						return DefWindowProc( hwnd, uMsg, wParam, lParam );
					break;
				default :
					return DefWindowProc( hwnd, uMsg, wParam, lParam );
			}
			break;

		case (UINT) WM_MSGBACKAGAIN :
			{
				if ( ( wParam == (WPARAM) 0 ) && ( lParam == (LPARAM) WM_LBUTTONDBLCLK ) )
					BackAgainApp( );
			}
			break;

		default :
			return DefWindowProc( hwnd, uMsg, wParam, lParam );
	}

	return TRUE;
}

BOOL CALLBACK CounterDetailsDlgProc( HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch ( uMsg )
	{
		case (UINT) WM_COMMAND :
			switch ( LOWORD( wParam ) )
			{
				case (WORD) IDOK :
				case (WORD) IDCANCEL :
					EndDialog( hwndDlg, wParam );
					break;
			}
			break;

		case (UINT) WM_INITDIALOG :
			{
				static _TCHAR szt[ 40 ];

				_stprintf( szt, TEXT( "%02u-%02u-%04u, %02u:%02u:%02u,%03u;" ),
					systimeStart.wDay, systimeStart.wMonth, systimeStart.wYear, systimeStart.wHour, systimeStart.wMinute, systimeStart.wSecond, systimeStart.wMilliseconds );
				SetDlgItemText( hwndDlg, IDC_RUNTIMEEDIT, szt );

				_stprintf( szt, TEXT( "%u times" ), uKbdKeyCount );
				SetDlgItemText( hwndDlg, IDC_KBDKEYSEDIT, szt );

				_stprintf( szt, TEXT( "%u times" ), uKbdSysKeyCount );
				SetDlgItemText( hwndDlg, IDC_KBDSYSKEYSEDIT, szt );

				_stprintf( szt, TEXT( "%u times" ), uMouseLeftCount );
				SetDlgItemText( hwndDlg, IDC_MOUSELEFTBTTNEDIT, szt );

				_stprintf( szt, TEXT( "%u times" ), uMouseRghtCount );
				SetDlgItemText( hwndDlg, IDC_MOUSERGHTBTTNEDIT, szt );

				if ( uMouseMiddleCount > 0 )
					_stprintf( szt, TEXT( "%u times" ), uMouseMiddleCount );
				else
					_tcscpy( szt, TEXT( "n/a" ) );
				SetDlgItemText( hwndDlg, IDC_MOUSEMIDDLEBTTNEDIT, szt );

				if ( uMouseOthrCount > 0 )
					_stprintf( szt, TEXT( "%u times" ), uMouseOthrCount );
				else
					_tcscpy( szt, TEXT( "n/a" ) );
				SetDlgItemText( hwndDlg, IDC_MOUSEOTHRBTTNEDIT, szt );

				if ( uMouseWheelCount > 0 )
					_stprintf( szt, TEXT( "%u times" ), uMouseWheelCount );
				else
					_tcscpy( szt, TEXT( "n/a" ) );
				SetDlgItemText( hwndDlg, IDC_MOUSEWHEELEDIT, szt );

				_stprintf( szt, TEXT( "%lu pixel(s)" ), ulMouseMoveDistance );
				SetDlgItemText( hwndDlg, IDC_MOUSEMOVEDISTANCEEDIT, szt );
			}
			break;

		default :
			return FALSE;
	}

	return TRUE;
}

static int __stdcall sgn( int n )
{
	return ( n > 0 ) ? 1 : ( n < 0 ) ? -1 : 0;
}

static int __stdcall line( int x0, int y0, int x1, int y1 )
{
	int dx =  abs( x1 - x0 ), sx = ( x0 < x1 ) ? 1 : -1;
	int dy = -abs( y1 - y0 ), sy = ( y0 < y1 ) ? 1 : -1;
	int err = dx + dy, e2; /* error value e_xy */
	int len = 0;

	while ( 1 )
	{
		len++;
		if ( ( x0 == x1 ) && ( y0 == y1 ) )
			break;
		e2 = 2*err;
		if ( e2 > dy ) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
		if ( e2 < dx ) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
	}

	return len;
}

unsigned long __stdcall CalcDistance( POINT& ptDest, const POINT& ptSrc )
{
	if ( ( ptSrc.x >= 0L ) && ( ptSrc.y >= 0L ) && ( ptSrc.x <= 10240L ) && ( ptSrc.y <= 10240L ) )
	{
		long nRet = line( (int) ptSrc.x, (int) ptSrc.y, (int) ptDest.x, (int) ptDest.y );

		ptDest.x = ptSrc.x;
		ptDest.y = ptSrc.y;

		return nRet;
	}
	else
		return 0L;
}

void __stdcall ResetCounter( void )
{
	GetLocalTime( (LPSYSTEMTIME) &systimeStart );

	uKbdKeyCount = 0U;
	uKbdSysKeyCount = 0U;
	uMouseLeftCount = 0U;
	uMouseRghtCount = 0U;
	uMouseMiddleCount = 0U;
	uMouseOthrCount = 0U;
	uMouseWheelCount = 0U;
	ulMouseMoveDistance = 0UL;
}

void __stdcall BackAgainApp( void )
{
	boolIsHidden = (BOOL) FALSE;

	NotificationSymbol( (DWORD) NIM_DELETE );

	ShowWindow( hwndMain, SW_RESTORE );
	UpdateWindow( hwndMain );
}

void __stdcall HideApp( void ) 
{
	ShowWindow( hwndMain, SW_HIDE );
	UpdateWindow( hwndMain );

	NotificationSymbol( (DWORD) NIM_ADD );

	boolIsHidden = (BOOL) TRUE;
}

void __stdcall NotificationSymbol( DWORD dwMsg )
{
	_TCHAR szt[ 80 ];

	_stprintf( szt, TEXT( "Event-Counter [%u keyboard events, %u mouse events]" ), uKbdKeyCount + uKbdSysKeyCount,
		uMouseLeftCount + uMouseRghtCount + uMouseMiddleCount + uMouseOthrCount + uMouseWheelCount );

	NOTIFYICONDATA notifyicondata;

	notifyicondata.cbSize = (DWORD) sizeof( NOTIFYICONDATA );
	notifyicondata.hWnd = (HWND) hwndMain;
	notifyicondata.uID = (UINT) 0;
	if ( dwMsg == (DWORD) NIM_DELETE )
	{
		notifyicondata.uFlags = (UINT) 0;
	}
	else
	{
		notifyicondata.uFlags = (UINT) ( NIF_ICON | NIF_MESSAGE | NIF_TIP );
		notifyicondata.uCallbackMessage = (UINT) WM_MSGBACKAGAIN;
		notifyicondata.hIcon = (HICON) hiconApp;
		lstrcpy( (LPTSTR) notifyicondata.szTip, (LPCTSTR) szt );
	}

	Shell_NotifyIcon( dwMsg, (PNOTIFYICONDATA) &notifyicondata );
}
