// evtlib.cxx

#define _WIN32_WINNT 0x0500
#include <windows.h>

#include "winnthook.h"

#pragma data_seg( ".SHARED" )
static int nInstanceCount = 0;
#pragma data_seg( )
#pragma comment(linker,"/section:.SHARED,RWS")

static HINSTANCE hInstance;

static HHOOK hKeybrdHook = (HHOOK) NULL;
static HHOOK hMouseHook = (HHOOK) NULL;

static void (CALLBACK * pfnKeybrdHookHlpr) ( WPARAM wParam, LPARAM lParam ) = (void (CALLBACK *) ( WPARAM, LPARAM )) NULL;
static void (CALLBACK * pfnMouseHookHlpr) ( WPARAM wParam, LPARAM lParam ) = (void (CALLBACK *) ( WPARAM, LPARAM )) NULL;

__declspec( dllexport ) int __stdcall InitHooks( void );
__declspec( dllexport ) int __stdcall DoneHooks( void );

__declspec( dllexport ) int __stdcall GetInstanceCount( void );

__declspec( dllexport ) VOID __stdcall SetKeybrdHookHlpr( void (CALLBACK * pfnKeybrdHookHlprArg) ( WPARAM, LPARAM ) );
__declspec( dllexport ) VOID __stdcall SetMouseHookHlpr( void (CALLBACK * pfnMouseHookHlprArg) ( WPARAM, LPARAM ) );

__declspec( dllexport ) LRESULT CALLBACK KeybrdHookProc( int code, WPARAM wParam, LPARAM lParam );
__declspec( dllexport ) LRESULT CALLBACK MouseHookProc( int code, WPARAM wParam, LPARAM lParam );

#ifdef __cplusplus
extern "C" {
#endif

int APIENTRY DllMain( HINSTANCE hInstance, DWORD dwReason, LPVOID /*lpReserved*/ )
{   
	switch ( dwReason )
	{
		case (DWORD) DLL_PROCESS_ATTACH :
			::hInstance = hInstance;
			nInstanceCount++;

		case (DWORD) DLL_THREAD_ATTACH :
			break;

		case (DWORD) DLL_PROCESS_DETACH :
			nInstanceCount--;

		case (DWORD) DLL_THREAD_DETACH :
			break;

		default :
			return 0;
	}

	return 1;
}

#ifdef __cplusplus
}
#endif

int __stdcall InitHooks( void )
{
	DoneHooks( );

	hKeybrdHook = SetWindowsHookEx( WH_KEYBOARD_LL, (HOOKPROC) KeybrdHookProc, (HINSTANCE) hInstance, 0 );
	hMouseHook = SetWindowsHookEx( WH_MOUSE_LL, (HOOKPROC) MouseHookProc, (HINSTANCE) hInstance, 0 );

	return (BOOL) hKeybrdHook && (BOOL) hMouseHook ? 0 : -1;
}

int __stdcall DoneHooks( void )
{
	if ( hKeybrdHook )
		UnhookWindowsHookEx( hKeybrdHook );
	if ( hMouseHook )
		UnhookWindowsHookEx( hMouseHook );

	hKeybrdHook = (HHOOK) NULL;
	hMouseHook = (HHOOK) NULL;

	return 0;
}

int __stdcall GetInstanceCount( void )
{
	return nInstanceCount;
}

VOID __stdcall SetKeybrdHookHlpr( void (CALLBACK * pfnKeybrdHookHlprArg) ( WPARAM, LPARAM ) )
{
	pfnKeybrdHookHlpr = pfnKeybrdHookHlprArg;
}

VOID __stdcall SetMouseHookHlpr( void (CALLBACK * pfnMouseHookHlprArg) ( WPARAM, LPARAM ) )
{
	pfnMouseHookHlpr = pfnMouseHookHlprArg;
}

LRESULT CALLBACK KeybrdHookProc( int code, WPARAM wParam, LPARAM lParam )
{
	if ( ( code == HC_ACTION ) || ( code == HC_NOREMOVE ) )
	{
		if ( pfnKeybrdHookHlpr )
			pfnKeybrdHookHlpr( wParam, lParam );
	}

	return CallNextHookEx( (HHOOK) NULL, code, wParam, lParam );
}

LRESULT CALLBACK MouseHookProc( int code, WPARAM wParam, LPARAM lParam )
{
	if ( ( code == HC_ACTION ) || ( code == HC_NOREMOVE ) )
	{
		if ( pfnMouseHookHlpr )
			pfnMouseHookHlpr( wParam, lParam );
	}

	return CallNextHookEx( (HHOOK) NULL, code, wParam, lParam );
}
